/*
	DrCr: Web-based double-entry bookkeeping framework
	Copyright (C) 2022-2025  Lee Yingtong Li (RunasSudo)

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU Affero General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Affero General Public License for more details.

	You should have received a copy of the GNU Affero General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::fmt::Display;

use async_trait::async_trait;
use mlua::{Function, Lua, LuaSerdeExt, Table, Value};
use serde::{Deserialize, Serialize};
use tokio::sync::RwLock;

use crate::reporting::calculator::ReportingGraphDependencies;
use crate::reporting::executor::ReportingExecutionError;
use crate::reporting::types::{
	ReportingContext, ReportingProductId, ReportingProductKind, ReportingProducts, ReportingStep,
	ReportingStepArgs, ReportingStepId,
};

fn load_plugin(plugin_dir: &str, plugin_name: &str) -> (Lua, Plugin) {
	let lua = Lua::new();

	// Init Lua environment
	let package = lua.globals().get::<Table>("package").unwrap();
	package
		.set("path", format!("{}/?.luau", plugin_dir))
		.unwrap();

	// Require and call the plugin
	let require = lua.load("require").eval::<Function>().unwrap();
	let plugin_table = require.call::<Table>(plugin_name).expect("Lua error");

	// Convert plugin to Rust struct
	let plugin = Plugin {
		spec: lua
			.from_value(
				plugin_table
					.get("spec")
					.expect("Error parsing Plugin definition"),
			)
			.unwrap(),
		requires: plugin_table
			.get("requires")
			.expect("Error parsing Plugin definition"),
		execute: plugin_table
			.get("execute")
			.expect("Error parsing Plugin definition"),
	};

	(lua, plugin)
}

/// Call [ReportingContext::register_lookup_fn] for all steps provided by this module
pub fn register_lookup_fns(context: &mut ReportingContext) {
	for plugin_path in context.plugin_names.clone().iter() {
		let (_, plugin) = load_plugin(&context.plugin_dir, plugin_path);

		for reporting_step in plugin.spec.reporting_steps.iter() {
			context.register_lookup_fn(
				reporting_step.name.clone(),
				reporting_step.product_kinds.clone(),
				PluginReportingStep::takes_args,
				PluginReportingStep::from_args,
			);
		}

		context
			.plugin_specs
			.insert(plugin_path.clone(), plugin.spec);
	}
}

#[derive(Debug)]
pub struct Plugin {
	spec: PluginSpec,
	requires: Function,
	execute: Function,
}

/// Represents a libdrcr plugin
#[derive(Debug, Deserialize, Serialize)]
pub struct PluginSpec {
	name: String,
	reporting_steps: Vec<ReportingStepSpec>,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct ReportingStepSpec {
	name: String,
	product_kinds: Vec<ReportingProductKind>,
}

/// Generic reporting step which is implemented by a plugin
#[derive(Debug)]
pub struct PluginReportingStep {
	pub plugin_path: String,
	pub step_name: String,
	pub product_kinds: Vec<ReportingProductKind>,
	pub args: ReportingStepArgs, // Currently only VoidArgs is supported
}

impl PluginReportingStep {
	fn takes_args(_name: &str, args: &ReportingStepArgs, _context: &ReportingContext) -> bool {
		*args == ReportingStepArgs::VoidArgs
	}

	fn from_args(
		name: &str,
		args: ReportingStepArgs,
		context: &ReportingContext,
	) -> Box<dyn ReportingStep> {
		// Look up plugin
		for (plugin_path, plugin_spec) in context.plugin_specs.iter() {
			if let Some(reporting_step_spec) =
				plugin_spec.reporting_steps.iter().find(|s| s.name == name)
			{
				return Box::new(Self {
					plugin_path: plugin_path.to_string(),
					step_name: name.to_string(),
					product_kinds: reporting_step_spec.product_kinds.clone(),
					args,
				});
			}
		}

		panic!("No plugin provides step {}", name);
	}
}

impl Display for PluginReportingStep {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		f.write_fmt(format_args!("{} {{PluginReportingStep}}", self.id()))
	}
}

#[async_trait]
impl ReportingStep for PluginReportingStep {
	fn id(&self) -> ReportingStepId {
		ReportingStepId {
			name: self.step_name.clone(),
			product_kinds: self.product_kinds.clone(),
			args: self.args.clone(),
		}
	}

	fn requires(&self, context: &ReportingContext) -> Vec<ReportingProductId> {
		// Call to plugin
		let (lua, plugin) = load_plugin(&context.plugin_dir, &self.plugin_path);
		let result_table = plugin
			.requires
			.call::<Table>((
				lua.to_value(&self.step_name).unwrap(),
				lua.to_value(&self.product_kinds).unwrap(),
				lua.to_value(&self.args).unwrap(),
			))
			.expect("Lua error");

		// Convert result to Rust
		let result = result_table
			.sequence_values()
			.map(|s| s.expect("Lua error"))
			.map(|v| lua.from_value(v).expect("Deserialise error"))
			.collect::<Vec<ReportingProductId>>();

		result
	}

	async fn execute(
		&self,
		context: &ReportingContext,
		_steps: &Vec<Box<dyn ReportingStep>>,
		_dependencies: &ReportingGraphDependencies,
		_products: &RwLock<ReportingProducts>,
	) -> Result<ReportingProducts, ReportingExecutionError> {
		// Call to plugin
		let (lua, plugin) = load_plugin(&context.plugin_dir, &self.plugin_path);
		let _result_table = plugin
			.execute
			.call::<Table>((
				lua.to_value(&self.step_name).unwrap(),
				lua.to_value(&self.product_kinds).unwrap(),
				lua.to_value(&self.args).unwrap(),
			))
			.expect("Lua error");

		eprintln!("Stub: Lua plugin execute");

		Ok(ReportingProducts::new())
	}
}

/// Format the [Table] as a string
fn _dbg_table(table: Table) -> String {
	format!(
		"{{{}}}",
		table
			.pairs::<Value, Value>()
			.map(|p| p.expect("Lua error"))
			.map(|(k, v)| format!("{:?}: {:?}", k, v))
			.collect::<Vec<_>>()
			.join(", ")
	)
}
